import base64
import hashlib
import io
import json
import logging
import re
import secrets
import uuid
from collections.abc import Callable
from textwrap import dedent
from typing import Any
from urllib.parse import parse_qs, urlencode

import httpx
from bs4 import BeautifulSoup, NavigableString, Tag
from PIL import Image

from .metadata import encrypt_metadata, meta_audible_app


logger = logging.getLogger("audible.login")

USER_AGENT = (
    "Mozilla/5.0 (iPhone; CPU iPhone OS 15_0 like Mac OS X) "
    "AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148"
)


def default_captcha_callback(captcha_url: str) -> str:
    """Helper function for handling captcha."""
    captcha = httpx.get(captcha_url).content
    f = io.BytesIO(captcha)
    img = Image.open(f)
    img.show()
    guess = input("Answer for CAPTCHA: ")
    return str(guess).strip().lower()


def default_otp_callback() -> str:
    """Helper function for handling 2-factor authentication."""
    guess = input("OTP Code: ")
    return str(guess).strip().lower()


def default_cvf_callback() -> str:
    """Helper function for handling cvf verifys.

    Amazon sends a verify code via Mail or SMS.
    """
    guess = input("CVF Code: ")
    return str(guess).strip().lower()


def default_approval_alert_callback() -> None:
    """Helper function for handling approval alerts."""
    print("Approval alert detected! Amazon sends you a mail.")
    input("Please press ENTER when you approve the notification.")


def playwright_external_login_url_callback(url: str) -> str:
    """Helper function for login using playwright."""
    from playwright.sync_api import sync_playwright  # type: ignore[import-not-found]  # noqa: I001

    with sync_playwright() as p:
        iphone = p.devices["iPhone 12 Pro"]
        browser = p.webkit.launch(headless=False)
        context = browser.new_context(**iphone)
        cookies = []
        for name, value in build_init_cookies().items():
            cookies.append({"name": name, "value": value, "url": url})
        context.add_cookies(cookies)
        page = browser.new_page()
        page.goto(url)

        while True:
            page.wait_for_timeout(600)
            if "/ap/maplanding" in page.url:
                response_url: str = page.url
                break

        browser.close()

    return response_url


def default_login_url_callback(url: str) -> str:
    """Helper function for login with external browsers."""
    try:
        return playwright_external_login_url_callback(url)
    except ImportError:
        pass

    message = f"""\
        Please copy the following url and insert it into a web browser of your choice:

        {url}

        Now you have to login with your Amazon credentials. After submit your username
        and password you have to do this a second time and solving a captcha before
        sending the login form.

        After login, your browser will show you an error page (Page not found). Do not
        worry about this. It has to be like this. Please copy the url from the address
        bar in your browser now.

        IMPORTANT:
        If you are using MacOS and have trouble insert the login result url, simply
        import the readline module in your script.

        Please insert the copied url (after login):
    """
    print(dedent(message))
    return input()


def _extract_message_from_box(box: Tag) -> str:
    message = ""

    header = box.find("h4")
    if isinstance(header, Tag) and header.string:
        message += header.string.strip()

    for list_item in box.find_all("li"):
        if isinstance(list_item, Tag) and isinstance(list_item.find("span"), Tag):
            list_entry = list_item.find("span")
            if isinstance(list_entry, Tag) and list_entry.string:
                message += " " + list_entry.string.strip()
    return message


def _get_messages_in_soup(soup: BeautifulSoup) -> dict[str, str]:
    messages = {}

    error_box = soup.find(id="auth-error-message-box")
    if isinstance(error_box, Tag):
        error_message = _extract_message_from_box(error_box)
        if error_message:
            messages["error"] = error_message

    warning_box = soup.find(id="auth-warning-message-box")
    if isinstance(warning_box, Tag):
        warning_message = _extract_message_from_box(warning_box)
        if warning_message:
            messages["warning"] = warning_message

    ap_error = soup.find(id="ap_error_page_message")
    if isinstance(ap_error, Tag):
        ap_error_message = ap_error.find(recursive=False, text=True)
        if isinstance(ap_error_message, NavigableString):
            messages["aperror"] = ap_error_message.strip()

    return messages


def get_soup(resp: httpx.Response, log_errors: bool = True) -> BeautifulSoup:
    soup = BeautifulSoup(resp.text, "html.parser")

    if log_errors:
        soup_messages = _get_messages_in_soup(soup)
        if "error" in soup_messages:
            logger.error("Error message: %s", soup_messages["error"])
        if "warning" in soup_messages:
            logger.warning("Warning message: %s", soup_messages["warning"])
        if "aperror" in soup_messages:
            logger.error("Error message: %s", soup_messages["aperror"])
    return soup


def get_inputs_from_soup(
    soup: BeautifulSoup, search_field: dict[str, str] | None = None
) -> dict[str, str]:
    """Extracts hidden form input fields from a Amazon login page."""
    search_field = search_field or {"name": "signIn"}
    form = soup.find("form", search_field) or soup.find("form")

    if not isinstance(form, Tag):
        raise Exception("No form found in page or something other is going wrong.")

    inputs = {}
    for field in form.find_all("input"):
        try:
            inputs[field["name"]] = ""
            if field["type"] and field["type"] == "hidden":
                inputs[field["name"]] = field["value"]
        except BaseException:  # noqa: S110
            pass
    return inputs


def get_next_action_from_soup(
    soup: BeautifulSoup, search_field: dict[str, str] | None = None
) -> tuple[str, str]:
    search_field = search_field or {"name": "signIn"}
    form = soup.find("form", search_field) or soup.find("form")

    if not isinstance(form, Tag):
        raise Exception("No form found in page or something other is going wrong.")

    method = form.get("method", "GET")
    url = form["action"]

    if not isinstance(method, str) or not isinstance(url, str):
        raise Exception("Error during extraction of next action in page.")

    return method, url


def create_code_verifier(length: int = 32) -> bytes:
    verifier = secrets.token_bytes(length)
    return base64.urlsafe_b64encode(verifier).rstrip(b"=")


def create_s256_code_challenge(verifier: bytes) -> bytes:
    m = hashlib.sha256(verifier)
    return base64.urlsafe_b64encode(m.digest()).rstrip(b"=")


def build_device_serial() -> str:
    return uuid.uuid4().hex.upper()


def build_client_id(serial: str) -> str:
    client_id = serial.encode() + b"#A2CZJZGLK2JJVM"
    return client_id.hex()


def build_oauth_url(
    country_code: str,
    domain: str,
    market_place_id: str,
    code_verifier: bytes,
    serial: str | None = None,
    with_username: bool = False,
) -> tuple[str, str]:
    """Builds the url to login to Amazon as an Audible device."""
    if with_username and domain.lower() not in ("de", "com", "co.uk"):
        raise ValueError(
            "Login with username is only supported for DE, US " "and UK marketplaces!"
        )

    serial = serial or build_device_serial()
    client_id = build_client_id(serial)
    code_challenge = create_s256_code_challenge(code_verifier)

    base_url = f"https://www.amazon.{domain}/ap/signin"
    return_to = f"https://www.amazon.{domain}/ap/maplanding"
    assoc_handle = f"amzn_audible_ios_{country_code}"
    page_id = "amzn_audible_ios"

    if with_username:
        base_url = f"https://www.audible.{domain}/ap/signin"
        return_to = f"https://www.audible.{domain}/ap/maplanding"
        assoc_handle = f"amzn_audible_ios_lap_{country_code}"
        page_id = "amzn_audible_ios_privatepool"

    oauth_params = {
        "openid.oa2.response_type": "code",
        "openid.oa2.code_challenge_method": "S256",
        "openid.oa2.code_challenge": code_challenge,
        "openid.return_to": return_to,
        "openid.assoc_handle": assoc_handle,
        "openid.identity": "http://specs.openid.net/auth/2.0/" "identifier_select",
        "pageId": page_id,
        "accountStatusPolicy": "P1",
        "openid.claimed_id": "http://specs.openid.net/auth/2.0/" "identifier_select",
        "openid.mode": "checkid_setup",
        "openid.ns.oa2": "http://www.amazon.com/ap/ext/oauth/2",
        "openid.oa2.client_id": f"device:{client_id}",
        "openid.ns.pape": "http://specs.openid.net/extensions/pape/1.0",
        "marketPlaceId": market_place_id,
        "openid.oa2.scope": "device_auth_access",
        "forceMobileLayout": "true",
        "openid.ns": "http://specs.openid.net/auth/2.0",
        "openid.pape.max_auth_age": "0",
    }

    return f"{base_url}?{urlencode(oauth_params)}", serial


def build_init_cookies() -> dict[str, str]:
    """Build initial cookies to prevent captcha in most cases."""
    token_bytes = secrets.token_bytes(313)
    frc = base64.b64encode(token_bytes).decode("ascii").rstrip("=")

    map_md_dict = {
        "device_user_dictionary": [],
        "device_registration_data": {"software_version": "35602678"},
        "app_identifier": {"app_version": "3.56.2", "bundle_id": "com.audible.iphone"},
    }
    map_md_str = json.dumps(map_md_dict)
    map_md = base64.b64encode(map_md_str.encode()).decode().rstrip("=")

    amzn_app_id = "MAPiOSLib/6.0/ToHideRetailLink"

    return {"frc": frc, "map-md": map_md, "amzn-app-id": amzn_app_id}


def check_for_captcha(soup: BeautifulSoup) -> bool:
    """Checks a Amazon login page for a captcha form."""
    captcha = soup.find("img", alt=lambda x: x and "CAPTCHA" in x)
    return True if captcha else False


def extract_captcha_url(soup: BeautifulSoup) -> str:
    """Returns the captcha url from a Amazon login page."""
    captcha = soup.find("img", alt=lambda x: x and "CAPTCHA" in x)
    if not isinstance(captcha, Tag) or not isinstance(captcha["src"], str):
        raise Exception("Error during extracting the captcha url.")
    return captcha["src"]


def check_for_mfa(soup: BeautifulSoup) -> bool:
    """Checks a Amazon login page for a multi-factor authentication form."""
    mfa = soup.find(
        "form",
        id=lambda x: x and ("verification-code-form" in x or "auth-mfa-form" in x),
    )
    return True if mfa else False


def check_for_choice_mfa(soup: BeautifulSoup) -> bool:
    """Checks a Amazon login page for a MFA selection form."""
    mfa_choice = soup.find("form", id="auth-select-device-form")
    return True if mfa_choice else False


def check_for_cvf(soup: BeautifulSoup) -> bool:
    cvf = soup.find("div", id="cvf-page-content")
    return True if cvf else False


def check_for_approval_alert(soup: BeautifulSoup) -> bool:
    """Checks a Amazon login page for an approval alert."""
    approval_alert = soup.find(id="resend-approval-alert") or soup.find(
        id="resend-approval-form"
    )
    return True if approval_alert else False


def extract_code_from_url(url: httpx.URL) -> str:
    """Extracts the access token from url query after login."""
    parsed_url = parse_qs(url.query.decode())
    return parsed_url["openid.oa2.authorization_code"][0]


def is_valid_email(obj: str) -> bool:
    valid_mail = r"^[a-z0-9]+[\._-]?[a-z0-9]+[@]\w+[.]\w+$"
    if re.match(valid_mail, obj):
        return True
    return False


def login(
    username: str,
    password: str,
    country_code: str,
    domain: str,
    market_place_id: str,
    serial: str | None = None,
    with_username: bool = False,
    captcha_callback: Callable[[str], str] | None = None,
    otp_callback: Callable[[], str] | None = None,
    cvf_callback: Callable[[], str] | None = None,
    approval_callback: Callable[[], Any] | None = None,
) -> dict[str, Any]:
    """Login to Audible by simulating an Audible App for iOS.

    Args:
        username: The Amazon email address.
        password: The Amazon password.
        country_code: The country code for the Audible marketplace to login.
        domain: domain: The top level domain for the Audible marketplace to
            login.
        market_place_id: The id for the Audible marketplace to login.
        serial: The device serial. If ``None`` a custom one will be created.
        with_username: If ``True`` login with Audible username instead
            of Amazon account.
        captcha_callback: A custom Callable for handling captcha requests.
            If ``None`` :func:`default_captcha_callback` is used.
        otp_callback: A custom Callable for providing one-time passwords.
            If ``None`` :func:`default_otp_callback` is used.
        cvf_callback: A custom Callable for providing the answer for a CVF
            code. If ``None`` :func:`default_cvf_callback` is used.
        approval_callback: A custom Callable for handling approval alerts.
            If ``None`` :func:`default_approval_alert_callback` is used.

    Returns:
        An ``authorization_code``, a ``code_verifier`` and the
        ``device serial`` from the authorized Client.

    Raises:
        Exception: If authorization_code is not in response url.
    """
    if with_username:
        base_url = f"https://www.audible.{domain}"
        logger.info("Login with Audible username.")
    else:
        if not is_valid_email(username):
            logger.warning("Username %s is not a valid mail address.", username)
        base_url = f"https://www.amazon.{domain}"
        logger.info("Login with Amazon Account.")

    default_headers = {
        "User-Agent": USER_AGENT,
        "Accept-Language": "en-US",
        "Accept-Encoding": "gzip",
    }
    init_cookies = build_init_cookies()

    session = httpx.Client(
        base_url=base_url,
        headers=default_headers,
        cookies=init_cookies,
        follow_redirects=True,
    )
    code_verifier = create_code_verifier()

    oauth_url, serial = build_oauth_url(
        country_code=country_code,
        domain=domain,
        market_place_id=market_place_id,
        code_verifier=code_verifier,
        serial=serial,
        with_username=with_username,
    )

    oauth_resp = session.get(oauth_url)
    oauth_soup = get_soup(oauth_resp)

    login_inputs = get_inputs_from_soup(oauth_soup)
    login_inputs["email"] = username
    login_inputs["password"] = password

    metadata = meta_audible_app(USER_AGENT, base_url)
    login_inputs["metadata1"] = encrypt_metadata(metadata)

    method, url = get_next_action_from_soup(oauth_soup, {"name": "signIn"})

    login_resp = session.request(method, url, data=login_inputs)
    login_soup = get_soup(login_resp)

    # check for captcha
    while check_for_captcha(login_soup):
        captcha_url = extract_captcha_url(login_soup)
        if captcha_callback:
            guess = captcha_callback(captcha_url)
        else:
            guess = default_captcha_callback(captcha_url)

        inputs = get_inputs_from_soup(login_soup)
        inputs["guess"] = guess
        inputs["use_image_captcha"] = "true"
        inputs["use_audio_captcha"] = "false"
        inputs["showPasswordChecked"] = "false"
        inputs["email"] = username
        inputs["password"] = password

        method, url = get_next_action_from_soup(login_soup, {"name": "signIn"})

        login_resp = session.request(method, url, data=inputs)
        login_soup = get_soup(login_resp)

    # check for choice mfa
    # https://www.amazon.de/ap/mfa/new-otp
    while check_for_choice_mfa(login_soup):
        inputs = get_inputs_from_soup(login_soup)
        for node in login_soup.select("div[data-a-input-name=otpDeviceContext]"):
            # auth-TOTP, auth-SMS, auth-VOICE
            if "auth-TOTP" in node["class"]:
                inp_node = node.find("input")
                if (
                    isinstance(inp_node, Tag)
                    and isinstance(inp_node["name"], str)
                    and isinstance(inp_node["value"], str)
                ):
                    inputs[inp_node["name"]] = inp_node["value"]

        method, url = get_next_action_from_soup(login_soup)

        login_resp = session.request(method, url, data=inputs)
        login_soup = get_soup(login_resp)

    # check for mfa (otp_code)
    while check_for_mfa(login_soup):
        if otp_callback:
            otp_code = otp_callback()
        else:
            otp_code = default_otp_callback()

        inputs = get_inputs_from_soup(login_soup)
        inputs["otpCode"] = otp_code
        inputs["mfaSubmit"] = "Submit"
        inputs["rememberDevice"] = "false"

        method, url = get_next_action_from_soup(login_soup)

        login_resp = session.request(method, url, data=inputs)
        login_soup = get_soup(login_resp)

    # check for cvf
    while check_for_cvf(login_soup):
        if cvf_callback:
            cvf_code = cvf_callback()
        else:
            cvf_code = default_cvf_callback()

        inputs = get_inputs_from_soup(login_soup)

        method, url = get_next_action_from_soup(login_soup)

        login_resp = session.request(method, url, data=inputs)
        login_soup = get_soup(login_resp)

        inputs = get_inputs_from_soup(login_soup)
        inputs["action"] = "code"
        inputs["code"] = cvf_code

        method, url = get_next_action_from_soup(login_soup)

        login_resp = session.request(method, url, data=inputs)
        login_soup = get_soup(login_resp)

    # check for approval alert
    while check_for_approval_alert(login_soup):
        if approval_callback:
            approval_callback()
        else:
            default_approval_alert_callback()

        # url = login_soup.find(id="resend-approval-link")["href"]
        url = str(login_resp.url)

        login_resp = session.get(url)
        login_soup = get_soup(login_resp)

        while login_soup.find(
            "span", {"class": "transaction-approval-word-break"}
        ):  # a-size-base-plus transaction-approval-word-break a-text-bold
            login_resp = session.get(url)
            login_soup = get_soup(login_resp)
            logger.info("still waiting for redirect")

    session.close()

    authcode_url = None
    if b"openid.oa2.authorization_code" in login_resp.url.query:
        authcode_url = login_resp.url
    elif len(login_resp.history) > 0:
        for history in login_resp.history:
            if b"openid.oa2.authorization_code" in history.url.query:
                authcode_url = history.url
                break

    if authcode_url is None:
        raise Exception("Login failed. Please check the log.")

    logger.debug("Login confirmed for %s", username)

    authorization_code = extract_code_from_url(authcode_url)

    return {
        "authorization_code": authorization_code,
        "code_verifier": code_verifier,
        "domain": domain,
        "serial": serial,
    }


def external_login(
    country_code: str,
    domain: str,
    market_place_id: str,
    serial: str | None = None,
    with_username: bool = False,
    login_url_callback: Callable[[str], str] | None = None,
) -> dict[str, Any]:
    """Gives the url to login with external browser and prompt for result.

    Note:
        If you are using MacOS and have trouble insert the login result url
        simply import the readline module in your script. See
        `#34 <https://github.com/mkb79/Audible/issues/34#issuecomment-766408640>`_.

    Args:
        country_code: The country code for the Audible marketplace to login.
        domain: The top level domain for the Audible marketplace to
            login.
        market_place_id: The id for the Audible marketplace to login.
        serial: The device serial. If ``None`` a custom one will be created.
        with_username: If ``True`` login with Audible username instead
            of Amazon account.
        login_url_callback: A custom Callable for handling login with external
            browsers. If ``None`` :func:`default_login_url_callback` is used.

    Returns:
        An ``authorization_code``, a ``code_verifier`` and the
        ``device serial`` from the authorized Client.
    """
    code_verifier = create_code_verifier()
    oauth_url, serial = build_oauth_url(
        country_code=country_code,
        domain=domain,
        market_place_id=market_place_id,
        code_verifier=code_verifier,
        serial=serial,
        with_username=with_username,
    )

    if login_url_callback:
        _response_url = login_url_callback(oauth_url)
    else:
        _response_url = default_login_url_callback(oauth_url)

    response_url = httpx.URL(_response_url)
    parsed_url = parse_qs(response_url.query.decode())

    authorization_code = parsed_url["openid.oa2.authorization_code"][0]

    return {
        "authorization_code": authorization_code,
        "code_verifier": code_verifier,
        "domain": domain,
        "serial": serial,
    }



import logging
import re

import httpx
from httpcore import ConnectError


logger = logging.getLogger("audible.localization")

LOCALE_TEMPLATES = {
    "germany": {
        "country_code": "de",
        "domain": "de",
        "market_place_id": "AN7V1F1VY261K",
    },
    "united_states": {
        "country_code": "us",
        "domain": "com",
        "market_place_id": "AF2M0KC94RCEA",
    },
    "united_kingdom": {
        "country_code": "uk",
        "domain": "co.uk",
        "market_place_id": "A2I9A3Q2GNFNGQ",
    },
    "france": {
        "country_code": "fr",
        "domain": "fr",
        "market_place_id": "A2728XDNODOQ8T",
    },
    "canada": {
        "country_code": "ca",
        "domain": "ca",
        "market_place_id": "A2CQZ5RBY40XE",
    },
    "italy": {
        "country_code": "it",
        "domain": "it",
        "market_place_id": "A2N7FU2W2BU2ZC",
    },
    "australia": {
        "country_code": "au",
        "domain": "com.au",
        "market_place_id": "AN7EY7DTAW63G",
    },
    "india": {
        "country_code": "in",
        "domain": "in",
        "market_place_id": "AJO3FBRUE6J4S",
    },
    "japan": {
        "country_code": "jp",
        "domain": "co.jp",
        "market_place_id": "A1QAP3MOU4173J",
    },
    "spain": {
        "country_code": "es",
        "domain": "es",
        "market_place_id": "ALMIKO4SZCSAR",
    },
    "brazil": {
        "country_code": "br",
        "domain": "com.br",
        "market_place_id": "A10J1VAYUDTYRN",
    },
}


def search_template(key: str, value: str) -> dict[str, str] | None:
    for country in LOCALE_TEMPLATES:
        locale = LOCALE_TEMPLATES[country]
        if locale[key] == value:
            logger.debug("found locale for %s", country)
            return locale

    logger.info("do not found %s in %s", value, key)
    return None


def autodetect_locale(domain: str) -> dict[str, str]:
    """Try to automatically detect correct settings for marketplace.

    Needs the top level domain of the audible page to continue with
    (e.g. co.uk, co.jp) and returns results found.

    Args:
        domain: The top level domain for the Audible marketplace to
            detect settings for (e.g. com).

    Returns:
        The settings for the found Audible marketplace.

    Raises:
        ConnectError: If site does not exist or network error raises.
        Exception: If marketplace or country code can't be found.
    """
    domain = domain.lstrip(".")
    site = f"https://www.audible.{domain}"
    params = {"ipRedirectOverride": True, "overrideBaseCountry": True}

    try:
        resp = httpx.get(site, params=params)
    except ConnectError as e:
        logger.warning("site %s does not exists or Network Error occurs", site)
        raise e

    marketplace_pattern = re.compile(r"ue_mid = \'(.*)\'")
    marketplace_search = re.search(marketplace_pattern, resp.text)
    if marketplace_search is None:
        raise Exception("can't find marketplace")
    market_place_id = marketplace_search.group(1)

    alias_pattern = re.compile(r"autocomplete_config.searchAlias = \"(.*)\"")
    alias_search = re.search(alias_pattern, resp.text)
    if alias_search is None:
        raise Exception("can't find country code")
    country_code = alias_search.group(1).split("-")[-1]

    return {
        "country_code": country_code,
        "domain": domain,
        "market_place_id": market_place_id,
    }


class Locale:
    """Adjustments for the different marketplaces who are provided by Audible.

    You can try to ``autodetect_locale`` if your marketplace is
    not in templates`.

    """

    def __init__(
        self,
        country_code: str | None = None,
        domain: str | None = None,
        market_place_id: str | None = None,
    ) -> None:
        if country_code is None or domain is None or market_place_id is None:
            locale = None
            if country_code:
                locale = search_template("country_code", country_code)
            elif domain:
                locale = search_template("domain", domain)

            if locale is None:
                raise Exception("can't find locale")

            country_code = country_code or locale["country_code"]
            domain = domain or locale["domain"]
            market_place_id = market_place_id or locale["market_place_id"]

        self._country_code = country_code
        self._domain = domain
        self._market_place_id = market_place_id

    def __repr__(self) -> str:
        return (
            f"Locale class for domain: {self.domain}, "
            f"marketplace: {self.market_place_id}"
        )

    def to_dict(self) -> dict[str, str]:
        return {
            "country_code": self.country_code,
            "domain": self.domain,
            "market_place_id": self.market_place_id,
        }

    @property
    def country_code(self) -> str:
        return self._country_code

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def market_place_id(self) -> str:
        return self._market_place_id



import inspect
import json
import logging
from abc import ABCMeta, abstractmethod
from collections.abc import Callable, Coroutine
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from types import TracebackType
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    overload,
)

import httpx
from httpx import URL
from httpx._models import HeaderTypes  # type: ignore[attr-defined]

from ._types import TrueFalseT
from .auth import Authenticator
from .exceptions import (
    BadRequest,
    NetworkError,
    NotFoundError,
    NotResponding,
    RatelimitError,
    RequestError,
    ServerError,
    Unauthorized,
    UnexpectedError,
)
from .localization import LOCALE_TEMPLATES, Locale


logger = logging.getLogger("audible.client")

ClientT = TypeVar("ClientT", httpx.AsyncClient, httpx.Client)

httpx_client_request_args = list(
    inspect.signature(httpx.Client.request).parameters.keys()
)


def default_response_callback(resp: httpx.Response) -> Any:
    raise_for_status(resp)
    return convert_response_content(resp)


def raise_for_status(resp: httpx.Response) -> None:
    try:
        resp.raise_for_status()
    except httpx.HTTPStatusError as e:
        code = resp.status_code
        data = convert_response_content(resp)
        if code == 400:
            raise BadRequest(resp, data) from None
        elif code in (401, 403):  # Unauthorized request - Invalid credentials
            raise Unauthorized(resp, data) from None
        elif code == 404:  # not found
            raise NotFoundError(resp, data) from None
        elif code == 429:
            raise RatelimitError(resp, data) from None
        elif code == 503:  # Maintainence
            raise ServerError(resp, data) from None
        else:
            raise UnexpectedError(resp, data) from e


def convert_response_content(resp: httpx.Response) -> Any:
    try:
        return resp.json()
    except json.JSONDecodeError:
        return resp.text


class BaseClient(Generic[ClientT], metaclass=ABCMeta):
    _API_URL_TEMP = "https://api.audible."
    _API_VERSION = "1.0"
    _REQUEST_LOG = "{method} {url} has received {text}, has returned {status}"

    def __init__(
        self,
        auth: Authenticator,
        country_code: str | None = None,
        headers: HeaderTypes | None = None,
        timeout: int = 10,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **session_kwargs: Any,
    ):
        locale = Locale(country_code.lower()) if country_code else auth.locale
        if not isinstance(locale, Locale):
            raise Exception("Authenticator has no `Locale` class set.")
        self._api_url = httpx.URL(self._API_URL_TEMP + locale.domain)

        default_headers = httpx.Headers(
            {
                "Accept": "application/json",
                "Accept-Charset": "utf-8",
                "Content-Type": "application/json",
            }
        )
        if headers is not None:
            default_headers.update(headers)

        self.session: ClientT = self._get_session(
            headers=default_headers, timeout=timeout, auth=auth, **session_kwargs
        )

        if response_callback is None:
            response_callback = default_response_callback
        self._response_callback = response_callback

    @abstractmethod
    def _get_session(self, *args: Any, **kwargs: Any) -> ClientT: ...

    @abstractmethod
    def _request(
        self,
        method: str,
        path: str,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: Any,
    ) -> Any: ...

    def switch_marketplace(self, country_code: str) -> None:
        locale = Locale(country_code.lower())
        self._api_url = URL(self._API_URL_TEMP + locale.domain)

    @property
    def marketplace(self) -> str:
        api_url = str(self._api_url)
        slice_len = len(self._API_URL_TEMP)
        domain = api_url[slice_len:]

        for country in LOCALE_TEMPLATES.values():
            if domain == country["domain"]:
                return country["country_code"]

        return "unknown"

    @property
    def auth(self) -> Authenticator:
        if not isinstance(self.session.auth, Authenticator):
            auth_type = type(self.session.auth)
            raise Exception(
                f"`session.auth` has type {auth_type}, expected `Authenticator`."
            )
        return self.session.auth

    def switch_user(
        self, auth: Authenticator, switch_to_default_marketplace: bool = False
    ) -> None:
        if switch_to_default_marketplace:
            if not isinstance(auth.locale, Locale):
                raise Exception("Authenticator has no `Locale` class set.")
            self.switch_marketplace(auth.locale.country_code)
        self.session.auth = auth

    def get_user_profile(self) -> dict[str, Any]:
        self.auth.refresh_access_token()
        return self.auth.user_profile()

    @property
    def user_name(self) -> str:
        user_profile = self.get_user_profile()
        if "name" not in user_profile:
            raise Exception("user profile has no key `name`.")

        user_name = user_profile["name"]
        if not isinstance(user_name, str):
            user_name_type = type(user_name)
            raise Exception(f"username has type {user_name_type}, expected `str`.")
        return user_name

    def _prepare_api_path(self, path: str) -> httpx.URL:
        if httpx.URL(path).is_absolute_url:
            return httpx.URL(path)

        if path.startswith("/"):
            path = path[1:]

        if not (path.startswith(self._API_VERSION) or path.startswith("0.0")):
            path = "/".join((self._API_VERSION, path))

        path = "/" + path
        path_bytes = path.encode()
        return self._api_url.copy_with(raw_path=path_bytes)

    @overload
    def raw_request(
        self: "BaseClient[httpx.Client]",
        method: str,
        url: str,
        *,
        stream: Literal[False] = ...,
        apply_auth_flow: bool = ...,
        apply_cookies: bool = ...,
        **kwargs: Any,
    ) -> httpx.Response: ...

    @overload
    def raw_request(
        self: "BaseClient[httpx.Client]",
        method: str,
        url: str,
        *,
        stream: Literal[True],
        apply_auth_flow: bool = ...,
        apply_cookies: bool = ...,
        **kwargs: Any,
    ) -> AbstractContextManager[httpx.Response]: ...

    @overload
    def raw_request(
        self: "BaseClient[httpx.AsyncClient]",
        method: str,
        url: str,
        *,
        stream: Literal[False] = ...,
        apply_auth_flow: bool = ...,
        apply_cookies: bool = ...,
        **kwargs: Any,
    ) -> Coroutine[Any, Any, httpx.Response]: ...

    @overload
    def raw_request(
        self: "BaseClient[httpx.AsyncClient]",
        method: str,
        url: str,
        *,
        stream: Literal[True],
        apply_auth_flow: bool = ...,
        apply_cookies: bool = ...,
        **kwargs: Any,
    ) -> AbstractAsyncContextManager[httpx.Response]: ...

    def raw_request(
        self,
        method: str,
        url: str,
        *,
        stream: TrueFalseT = False,
        apply_auth_flow: bool = False,
        apply_cookies: bool = False,
        **kwargs: Any,
    ) -> httpx.Response | (
        Coroutine[Any, Any, httpx.Response]
        | (
            AbstractContextManager[httpx.Response]
            | AbstractAsyncContextManager[httpx.Response]
        )
    ):
        """Sends a raw request with the underlying httpx Client.

        This method ignores a set api_url and allows send request to custom
        hosts. The raw httpx response will be returned.

        Args:
            method: The http request method.
            url: The url to make requests to.
            stream: If `True`, streams the response
            apply_auth_flow: If `True`, the :meth:`Authenticator.auth_flow`
                will be applied to the request.
            apply_cookies: If `True`, website cookies from
                :attr:`Authenticator.website_cookies` will be added to
                request headers.
            **kwargs: keyword args supported by :class:`httpx.AsyncClient.stream`,
                :class:`httpx.Client.stream`, :class:`httpx.AsyncClient.request`,
                :class:`httpx.Client.request`.

        Returns:
            An unprepared httpx Response object.

        .. versionadded:: v0.5.1
        """
        request_params = {"method": method, "url": url, **kwargs}

        if apply_cookies:
            cookies = httpx.Cookies(self.auth.website_cookies)
            cookies.update(kwargs.pop("cookies", {}))
            request_params["cookies"] = cookies

        if apply_auth_flow:
            request_params["auth"] = self.auth

        if stream:
            return self.session.stream(**request_params)
        return self.session.request(**request_params)

    @staticmethod
    def _prepare_params(kwargs: dict[str, Any]) -> None:
        params = kwargs.pop("params", {})
        for key in list(kwargs.keys()):
            if key not in httpx_client_request_args:
                params[key] = kwargs.pop(key)
        kwargs["params"] = params

    @abstractmethod
    def get(
        self,
        path: str,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any: ...

    @abstractmethod
    def post(
        self,
        path: str,
        body: Any,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any: ...

    @abstractmethod
    def delete(
        self,
        path: str,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any: ...

    @abstractmethod
    def put(
        self,
        path: str,
        body: Any,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any: ...


class Client(BaseClient[httpx.Client]):
    def _get_session(self, *args: Any, **kwargs: Any) -> httpx.Client:
        return httpx.Client(*args, **kwargs)

    def __enter__(self) -> "Client":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self.close()

    def __repr__(self) -> str:
        return f"<Sync Client for *{self.marketplace}* marketplace>"

    def close(self) -> None:
        self.session.close()

    def _request(
        self,
        method: str,
        path: str,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        url = self._prepare_api_path(path)

        if response_callback is None:
            response_callback = self._response_callback

        try:
            resp = self.session.request(method, url, **kwargs)

            logger.debug(
                self._REQUEST_LOG.format(
                    method=method, url=resp.url, text=resp.text, status=resp.status_code
                )
            )

            return response_callback(resp)

        except (
            httpx.ConnectTimeout,
            httpx.ReadTimeout,
            httpx.WriteTimeout,
            httpx.PoolTimeout,
        ):
            raise NotResponding from None
        except httpx.NetworkError:
            raise NetworkError from None
        except httpx.RequestError as exc:
            raise RequestError(exc) from None
        finally:
            try:
                resp.close()
            except UnboundLocalError:
                pass

    def get(
        self,
        path: str,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any:
        self._prepare_params(kwargs)
        return self._request(
            method="GET", path=path, response_callback=response_callback, **kwargs
        )

    def post(
        self,
        path: str,
        body: Any,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any:
        self._prepare_params(kwargs)
        return self._request(
            method="POST",
            path=path,
            response_callback=response_callback,
            json=body,
            **kwargs,
        )

    def delete(
        self,
        path: str,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any:
        self._prepare_params(kwargs)
        return self._request(
            method="DELETE", path=path, response_callback=response_callback, **kwargs
        )

    def put(
        self,
        path: str,
        body: Any,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any:
        self._prepare_params(kwargs)
        return self._request(
            method="PUT",
            path=path,
            response_callback=response_callback,
            json=body,
            **kwargs,
        )


class AsyncClient(BaseClient[httpx.AsyncClient]):
    def _get_session(self, *args: Any, **kwargs: Any) -> httpx.AsyncClient:
        return httpx.AsyncClient(*args, **kwargs)

    async def __aenter__(self) -> "AsyncClient":
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        await self.close()

    def __repr__(self) -> str:
        return f"<AyncClient for *{self.marketplace}* marketplace>"

    async def close(self) -> None:
        await self.session.aclose()

    async def _request(
        self,
        method: str,
        path: str,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        url = self._prepare_api_path(path)

        if response_callback is None:
            response_callback = self._response_callback

        try:
            resp = await self.session.request(method, url, **kwargs)

            logger.debug(
                self._REQUEST_LOG.format(
                    method=method, url=resp.url, text=resp.text, status=resp.status_code
                )
            )

            return response_callback(resp)

        except (
            httpx.ConnectTimeout,
            httpx.ReadTimeout,
            httpx.WriteTimeout,
            httpx.PoolTimeout,
        ):
            raise NotResponding from None
        except httpx.NetworkError:
            raise NetworkError from None
        except httpx.RequestError as exc:
            raise RequestError(exc) from None
        finally:
            try:
                await resp.aclose()
            except UnboundLocalError:
                pass

    async def get(
        self,
        path: str,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any:
        self._prepare_params(kwargs)
        return await self._request(
            method="GET", path=path, response_callback=response_callback, **kwargs
        )

    async def post(
        self,
        path: str,
        body: Any,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any:
        self._prepare_params(kwargs)
        return await self._request(
            method="POST",
            path=path,
            response_callback=response_callback,
            json=body,
            **kwargs,
        )

    async def delete(
        self,
        path: str,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any:
        self._prepare_params(kwargs)
        return await self._request(
            method="DELETE", path=path, response_callback=response_callback, **kwargs
        )

    async def put(
        self,
        path: str,
        body: Any,
        response_callback: Callable[[httpx.Response], Any] | None = None,
        **kwargs: dict[str, Any],
    ) -> Any:
        self._prepare_params(kwargs)
        return await self._request(
            method="PUT",
            path=path,
            response_callback=response_callback,
            json=body,
            **kwargs,
        )



from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    import httpx


class RequestError(Exception):
    """Base class for all errors."""


class StatusError(RequestError):
    """Base class for all errors except NotResponding and RatelimitDetectedError."""

    def __init__(self, resp: "httpx.Response", data: Any) -> None:
        self.response = resp
        self.code = resp.status_code
        self.method = getattr(resp, "method", None)
        self.reason = resp.reason_phrase
        if isinstance(data, dict):
            self.error = data.get("error")
            if "message" in data:
                self.error = data.get("message")
        else:
            self.error = data
        self.fmt = f"{self.reason} ({self.code}): {self.error}"
        super().__init__(self.fmt)


class NotResponding(RequestError):
    """Raised if the API request timed out."""

    def __init__(self) -> None:
        self.code = 504
        self.error = "API request timed out, please be patient."
        super().__init__(self.error)


class NetworkError(RequestError):
    """Raised if there is an issue with the network (i.e. requests.ConnectionError)."""

    def __init__(self) -> None:
        self.code = 503
        self.error = "Network down."
        super().__init__(self.error)


class BadRequest(StatusError):
    """Raised when status code 400 is returned.

    Typically, when at least one search parameter was not provided.
    """


class NotFoundError(StatusError):
    """Raised if no result is found."""


class ServerError(StatusError):
    """Raised if the api service is having issues."""


class Unauthorized(StatusError):
    """Raised if you passed invalid credentials."""


class RatelimitError(StatusError):
    """Raised if ratelimit is hit."""


class UnexpectedError(StatusError):
    """Raised if the error was not caught."""


class AuthFlowError(Exception):
    """Raised if no auth method available."""


class NoRefreshToken(Exception):
    """Raised if refresh token is needed but not provided."""


class FileEncryptionError(Exception):
    """Raised if something is wrong with file encryption."""



__version__ = "0.10.0"

from ._logging import log_helper
from .auth import Authenticator
from .client import AsyncClient, Client


__all__ = ["__version__", "Authenticator", "log_helper", "Client", "AsyncClient"]



import base64
import json
import logging
from collections.abc import Callable, Generator, Iterator
from datetime import datetime, timedelta, timezone
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Optional,
    Union,
    cast,
    overload,
)

import httpx
import rsa
from httpx import Cookies

from .activation_bytes import get_activation_bytes as get_ab
from .aescipher import AESCipher, detect_file_encryption
from .exceptions import AuthFlowError, FileEncryptionError, NoRefreshToken
from .login import external_login, login
from .register import deregister as deregister_
from .register import register as register_
from .utils import test_convert


if TYPE_CHECKING:
    import pathlib

    from ._types import TrueFalseT
    from .localization import Locale


logger = logging.getLogger("audible.auth")


def refresh_access_token(
    refresh_token: str, domain: str, with_username: bool = False
) -> dict[str, Any]:
    """Refreshes an access token.

    Args:
        refresh_token: The refresh token obtained after a device
            registration.
        domain: The top level domain of the requested Amazon server
            (e.g. com).
        with_username: If ``True`` uses `audible` domain instead of `amazon`.

    Returns:
        A dict with the new access token and expiration timestamp.

    Note:
        The new access token is valid for 60 minutes.

    .. versionadded:: v0.8
        The with_username argument
    """
    body = {
        "app_name": "Audible",
        "app_version": "3.56.2",
        "source_token": refresh_token,
        "requested_token_type": "access_token",
        "source_token_type": "refresh_token",
    }

    target_domain = "audible" if with_username else "amazon"

    resp = httpx.post(f"https://api.{target_domain}.{domain}/auth/token", data=body)
    resp.raise_for_status()
    resp_dict = resp.json()

    expires_in_sec = int(resp_dict["expires_in"])
    expires = (
        datetime.now(timezone.utc) + timedelta(seconds=expires_in_sec)
    ).timestamp()

    return {"access_token": resp_dict["access_token"], "expires": expires}


def refresh_website_cookies(
    refresh_token: str, domain: str, cookies_domain: str, with_username: bool = False
) -> dict[str, str]:
    """Fetches website cookies for a specific domain.

    Args:
        refresh_token: The refresh token obtained after a device
            registration.
        domain: The top level domain of the requested Amazon server
            (e.g. com, de, fr).
        cookies_domain: The top level domain scope for the cookies
            (e.g. com, de, fr).
        with_username: If ``True`` uses `audible` domain instead of `amazon`.

    Returns:
        The requested cookies for the Amazon and Audible website for the given
        `cookies_domain` scope.

    .. versionadded:: v0.8
        The with_username argument
    """
    target_domain = "audible" if with_username else "amazon"

    url = f"https://www.{target_domain}.{domain}/ap/exchangetoken/cookies"

    body = {
        "app_name": "Audible",
        "app_version": "3.56.2",
        "source_token": refresh_token,
        "requested_token_type": "auth_cookies",
        "source_token_type": "refresh_token",
        "domain": f".{target_domain}.{cookies_domain}",
    }

    resp = httpx.post(url, data=body)
    resp.raise_for_status()
    resp_dict = resp.json()

    raw_cookies = resp_dict["response"]["tokens"]["cookies"]
    website_cookies = {}
    for domain_cookies in raw_cookies:
        for cookie in raw_cookies[domain_cookies]:
            website_cookies[cookie["Name"]] = cookie["Value"].replace(r'"', r"")

    return website_cookies


def user_profile(access_token: str, domain: str) -> dict[str, Any]:
    """Returns user profile from Amazon.

    Args:
        access_token: The valid access token for authentication.
        domain: The top level domain of the requested Amazon server
            (e.g. com, de, fr).

    Returns:
        The Amazon user profile for the authenticated user.

    Raises:
        Exception: If the user profile is malformed
    """
    headers = {"Authorization": f"Bearer {access_token}"}

    resp = httpx.get(f"https://api.amazon.{domain}/user/profile", headers=headers)
    resp.raise_for_status()
    profile = resp.json()

    if not isinstance(profile, dict) and "user_id" not in profile:
        raise Exception("Malformed user profile response.")

    return profile


def user_profile_audible(access_token: str, domain: str) -> dict[str, Any]:
    """Returns user profile from Audible.

    Args:
        access_token: The valid access token for authentication.
        domain: The top level domain of the requested Audible server
            (e.g. com, de, fr).

    Returns:
        The Audible user profile for the authenticated user.

    Raises:
        Exception: If the user profile is malformed
    """
    headers = {"Authorization": f"Bearer {access_token}"}

    resp = httpx.get(f"https://api.audible.{domain}/user/profile", headers=headers)
    resp.raise_for_status()
    profile = resp.json()

    if not isinstance(profile, dict) and "user_id" not in profile:
        raise Exception("Malformed user profile response.")

    return profile


def sign_request(
    method: str, path: str, body: bytes, adp_token: str, private_key: str
) -> dict[str, str]:
    """Helper function who creates signed headers for http requests.

    Args:
        path: The requested http url path and query.
        method: The http request method (GET, POST, DELETE, ...).
        body: The http message body.
        adp_token: The adp token obtained after a device registration.
        private_key: The rsa key obtained after device registration.

    Returns:
        A dict with the signed headers.
    """
    date = datetime.now(timezone.utc).isoformat("T") + "Z"
    str_body = body.decode("utf-8")

    data = f"{method}\n{path}\n{date}\n{str_body}\n{adp_token}"

    key = rsa.PrivateKey.load_pkcs1(private_key.encode("utf-8"))
    cipher = rsa.pkcs1.sign(data.encode(), key, "SHA-256")
    signed_encoded = base64.b64encode(cipher)

    signature = f"{signed_encoded.decode()}:{date}"

    return {
        "x-adp-token": adp_token,
        "x-adp-alg": "SHA256withRSA:1.0",
        "x-adp-signature": signature,
    }


class Authenticator(httpx.Auth):
    """Audible Authenticator class.

    Note:
        A new class instance have to be instantiated with
        :meth:`Authenticator.from_login` or :meth:`Authenticator.from_file`.

    .. versionadded:: v0.8
        The with_username attribute.

    Note:
        Auth data saved with v0.8 or later
        can not be loaded with versions less than v0.8!
        If an auth file for a pre-Amazon account (with_username=True) was
        created with v0.7.1 or v0.7.2 set `auth.with_username` to `True` and
        save the data again. After this deregistration, refreshing access
        tokens and requesting cookies for another domain will work for
        pre-Amazon accounts.
    """

    access_token: str | None = None
    activation_bytes: str | None = None
    adp_token: str | None = None
    crypter: AESCipher | None = None
    customer_info: dict[str, Any] | None = None
    device_info: dict[str, Any] | None = None
    device_private_key: str | None = None
    encryption: str | bool | None = None
    expires: float | None = None
    filename: Optional["pathlib.Path"] = None
    locale: Optional["Locale"] = None
    refresh_token: str | None = None
    store_authentication_cookie: dict[str, Any] | None = None
    website_cookies: dict[str, Any] | None = None
    with_username: bool | None = False
    requires_request_body: bool = True
    _forbid_new_attrs: bool = True
    _apply_test_convert: bool = True

    def __setattr__(self, attr: str, value: Any) -> None:
        if self._forbid_new_attrs and not hasattr(self, attr):
            msg = (
                f"{self.__class__.__name__} is frozen, can't " f"add attribute: {attr}."
            )
            logger.error(msg)
            raise AttributeError(msg)

        if self._apply_test_convert:
            value = test_convert(attr, value)
        object.__setattr__(self, attr, value)

    def __iter__(self) -> Iterator[str]:
        for i in self.__dict__:
            if self.__dict__[i] is not None and not i.startswith("_"):
                yield i

    def __len__(self) -> int:
        return len(list(iter(self)))

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__dict__})"

    def _update_attrs(self, **kwargs: Any) -> None:
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    @classmethod
    def from_dict(
        cls, data: dict[str, Any], locale: Union[str, "Locale"] | None = None
    ) -> "Authenticator":
        """Instantiate an Authenticator from authentication file.

        .. versionadded:: v0.7.1

        Args:
            data: A dictionary with the authentication data
            locale: The country code of the Audible marketplace to interact
                with. If ``None`` the country code from file is used.

        Returns:
            A new Authenticator instance.
        """
        auth = cls()

        locale_code: str | None = data.pop("locale_code", None)
        auth.locale = cast("Locale", locale or locale_code)

        if "login_cookies" in data:
            auth.website_cookies = data.pop("login_cookies")

        auth._update_attrs(**data)

        logger.info("load data from dictionary for locale %s", auth.locale.country_code)

        return auth

    @classmethod
    def from_file(
        cls,
        filename: Union[str, "pathlib.Path"],
        password: str | None = None,
        locale: Union[str, "Locale"] | None = None,
        encryption: bool | str | None = None,
        **kwargs: Any,
    ) -> "Authenticator":
        """Instantiate an Authenticator from authentication file.

        .. versionadded:: v0.5.0

        Args:
            filename:
                The name of the file with the authentication data.
            password: The password of the authentication file.
            locale: The country code of the Audible marketplace to interact
                with. If ``None`` the country code from file is used.
            encryption: The encryption style to use. Can be ``json`` or
                ``bytes``. If ``None``, encryption will be auto detected.
            **kwargs: Keyword arguments are passed to the
                :class:`~audible.aescipher.AESCipher` class. See below.

        Keyword Arguments:
            key_size (int, Optional):
            salt_marker (Optional[bytes]):
            kdf_iterations (:obj:`int`, optional):
            hashmod:
            mac:

        Returns:
            A new Authenticator instance.

        Raises:
            FileEncryptionError: If file ist encrypted without providing a password
        """
        auth = cls()
        auth.filename = cast("pathlib.Path", filename)
        auth.encryption = encryption or detect_file_encryption(auth.filename)

        if isinstance(auth.encryption, str):
            if password is None:
                message = "File is encrypted but no password provided."
                logger.critical(message)
                raise FileEncryptionError(message)
            auth.crypter = AESCipher(password, **kwargs)
            file_data = auth.crypter.from_file(auth.filename, auth.encryption)
        else:
            file_data = auth.filename.read_text()

        json_data = json.loads(file_data)

        locale_code = json_data.pop("locale_code", None)
        locale = locale or locale_code
        auth.locale = cast("Locale", locale)

        # login cookies where renamed to website cookies
        # old names must be adjusted
        if "login_cookies" in json_data:
            auth.website_cookies = json_data.pop("login_cookies")

        auth._update_attrs(**json_data)

        logger.info(
            "load data from file %s for locale %s",
            auth.filename,
            auth.locale.country_code,
        )
        return auth

    @classmethod
    def from_login(
        cls,
        username: str,
        password: str,
        locale: Union[str, "Locale"],
        serial: str | None = None,
        with_username: bool = False,
        captcha_callback: Callable[[str], str] | None = None,
        otp_callback: Callable[[], str] | None = None,
        cvf_callback: Callable[[], str] | None = None,
        approval_callback: Callable[[], Any] | None = None,
    ) -> "Authenticator":
        """Instantiate a new Authenticator with authentication data from login.

        .. versionadded:: v0.5.0

        .. versionadded:: v0.5.4
           The serial argument
           The with_username argument

        Args:
            username: The Amazon email address.
            password: The Amazon password.
            locale: The ``country_code`` or :class:`audible.localization.Locale`
                instance for the marketplace to login.
            serial: The device serial. If ``None`` a custom one will be created.
            with_username: If ``True`` login with Audible username instead
                of Amazon account.
            captcha_callback: A custom callback to handle captcha requests
                during login.
            otp_callback: A custom callback to handle one-time password
                requests during login.
            cvf_callback: A custom callback to handle verify code requests
                during login.
            approval_callback: A custom Callable for handling approval alerts.

        Returns:
            An :class:`~audible.auth.Authenticator` instance.
        """
        auth = cls()
        auth.locale = cast("Locale", locale)

        login_device = login(
            username=username,
            password=password,
            country_code=auth.locale.country_code,
            domain=auth.locale.domain,
            market_place_id=auth.locale.market_place_id,
            serial=serial,
            with_username=with_username,
            captcha_callback=captcha_callback,
            otp_callback=otp_callback,
            cvf_callback=cvf_callback,
            approval_callback=approval_callback,
        )
        logger.info("logged in to Audible as %s", username)
        register_device = register_(with_username=with_username, **login_device)

        auth._update_attrs(with_username=with_username, **register_device)
        logger.info("registered Audible device")

        return auth

    @classmethod
    def from_login_external(
        cls,
        locale: Union[str, "Locale"],
        serial: str | None = None,
        with_username: bool = False,
        login_url_callback: Callable[[str], str] | None = None,
    ) -> "Authenticator":
        """Instantiate a new Authenticator from login with external browser.

        .. versionadded:: v0.5.1

        .. versionadded:: v0.5.4
           The serial argument
           The with_username argument

        Args:
            locale: The ``country_code`` or :class:`audible.localization.Locale`
                instance for the marketplace to login.
            serial: The device serial. If ``None`` a custom one will be created.
            with_username: If ``True`` login with Audible username instead
                of Amazon account.
            login_url_callback: A custom Callable for handling login with
                external browsers.

        Returns:
            An :class:`~audible.auth.Authenticator` instance.
        """
        auth = cls()
        auth.locale = cast("Locale", locale)

        login_device = external_login(
            country_code=auth.locale.country_code,
            domain=auth.locale.domain,
            market_place_id=auth.locale.market_place_id,
            serial=serial,
            with_username=with_username,
            login_url_callback=login_url_callback,
        )
        logger.info("logged in to Audible.")

        register_device = register_(with_username=with_username, **login_device)

        auth._update_attrs(with_username=with_username, **register_device)
        logger.info("registered Audible device")

        return auth

    def auth_flow(
        self, request: httpx.Request
    ) -> Generator[httpx.Request, httpx.Response, None]:
        """Auth flow to be executed on every request by :mod:`httpx`.

        Args:
            request: The request made by ``httpx``.

        Yields:
            The next request

        Raises:
            AuthFlowError: If no auth flow is available.
        """
        available_modes = self.available_auth_modes

        if "signing" in available_modes:
            self._apply_signing_auth_flow(request)
        elif "bearer" in available_modes:
            self._apply_bearer_auth_flow(request)
        else:
            message = "signing or bearer auth flow are not available."
            logger.critical(message)
            raise AuthFlowError(message)

        yield request

    def _apply_signing_auth_flow(self, request: httpx.Request) -> None:
        if self.adp_token is None or self.device_private_key is None:
            raise Exception("No signing data found.")

        headers = sign_request(
            method=request.method,
            path=request.url.raw_path.decode(),
            body=request.content,
            adp_token=self.adp_token,
            private_key=self.device_private_key,
        )

        request.headers.update(headers)
        logger.info("signing auth flow applied to request")

    def _apply_bearer_auth_flow(self, request: httpx.Request) -> None:
        if self.access_token_expired:
            self.refresh_access_token()

        if self.access_token is None:
            raise Exception("No access token found.")

        headers = {"Authorization": "Bearer " + self.access_token, "client-id": "0"}
        request.headers.update(headers)
        logger.info("bearer auth flow applied to request")

    def _apply_cookies_auth_flow(self, request: httpx.Request) -> None:
        if self.website_cookies is None:
            raise Exception("No website cookies found.")
        cookies = self.website_cookies.copy()

        Cookies(cookies).set_cookie_header(request)
        logger.info("cookies auth flow applied to request")

    def sign_request(self, request: httpx.Request) -> None:
        """Sign a request.

        .. deprecated:: 0.5.0
            Use :meth:`self._apply_signing_auth_flow` instead.
        """
        self._apply_signing_auth_flow(request)

    @property
    def available_auth_modes(self) -> list[str]:
        available_modes = []

        if self.adp_token and self.device_private_key:
            available_modes.append("signing")

        if self.access_token:
            if self.access_token_expired and not self.refresh_token:
                pass
            else:
                available_modes.append("bearer")

        if self.website_cookies:
            available_modes.append("cookies")

        return available_modes

    def to_dict(self) -> dict[str, Any]:
        """Returns authentication data as dict.

        .. versionadded:: 0.7.1

        .. versionadded:: v0.8
           The returned dict now contains the `with_username` attribute
        """
        data = {
            "website_cookies": self.website_cookies,
            "adp_token": self.adp_token,
            "access_token": self.access_token,
            "refresh_token": self.refresh_token,
            "device_private_key": self.device_private_key,
            "store_authentication_cookie": self.store_authentication_cookie,
            "device_info": self.device_info,
            "customer_info": self.customer_info,
            "expires": self.expires,
            "locale_code": self.locale.country_code if self.locale else None,
            "with_username": self.with_username,
            "activation_bytes": self.activation_bytes,
        }
        return data

    def to_file(
        self,
        filename: Union["pathlib.Path", str] | None = None,
        password: str | None = None,
        encryption: bool | str = "default",
        indent: int = 4,
        set_default: bool = True,
        **kwargs: Any,
    ) -> None:
        """Save authentication data to file.

        .. versionadded:: 0.5.1
           Save activation bytes to auth file

        .. versionadded:: v0.8
           The saved file now contains the `with_username` attribute
        """
        if not (filename or self.filename):
            raise ValueError("No filename provided")

        if filename:
            target_file: pathlib.Path = test_convert("filename", filename)
        elif self.filename:
            target_file = self.filename
        else:
            raise ValueError("No filename provided")

        if encryption != "default":
            encryption = test_convert("encryption", encryption)
        else:
            encryption = self.encryption or False

        data = self.to_dict()
        json_data = json.dumps(data, indent=indent)

        if encryption is False:
            target_file.write_text(json_data)
            crypter = None
        else:
            if password:
                crypter = test_convert("crypter", AESCipher(password, **kwargs))
            elif self.crypter:
                crypter = self.crypter
            else:
                raise ValueError("No password provided")

            crypter.to_file(
                json_data, filename=target_file, encryption=encryption, indent=indent
            )

        logger.info("saved data to file %s", target_file)

        if set_default:
            self.filename = target_file
            self.encryption = encryption
            self.crypter = crypter

        logger.info("set filename %s as default", target_file)

    def deregister_device(self, deregister_all: bool = False) -> Any:
        self.refresh_access_token()

        if self.access_token is None:
            raise Exception("No access token found.")
        if self.locale is None:
            raise Exception("No locale found.")

        return deregister_(
            access_token=self.access_token,
            deregister_all=deregister_all,
            domain=self.locale.domain,
            with_username=self.with_username or False,
        )

    def refresh_access_token(self, force: bool = False) -> None:
        if force or self.access_token_expired:
            if self.refresh_token is None:
                message = "No refresh token found. Can't refresh access token."
                logger.critical(message)
                raise NoRefreshToken(message)
            if self.locale is None:
                raise Exception("No locale found.")

            refresh_data = refresh_access_token(
                refresh_token=self.refresh_token,
                domain=self.locale.domain,
                with_username=self.with_username or False,
            )

            self._update_attrs(**refresh_data)
        else:
            logger.info(
                "Access Token not expired. No refresh necessary. "
                "To force refresh please use force=True"
            )

    def set_website_cookies_for_country(self, country_code: str) -> None:
        cookies_domain = test_convert("locale", country_code).domain

        if self.refresh_token is None:
            raise Exception("No refresh token found.")
        if self.locale is None:
            raise Exception("No locale found.")

        self.website_cookies = refresh_website_cookies(
            self.refresh_token,
            self.locale.domain,
            cookies_domain,
            self.with_username or False,
        )

    @overload
    def get_activation_bytes(
        self,
        filename: Union["pathlib.Path", str] | None = ...,
        extract: Literal[True] = ...,
        force_refresh: bool = ...,
    ) -> str: ...

    @overload
    def get_activation_bytes(
        self,
        filename: Union["pathlib.Path", str] | None = ...,
        *,
        extract: Literal[False],
        force_refresh: bool = ...,
    ) -> bytes: ...

    def get_activation_bytes(
        self,
        filename: Union["pathlib.Path", str] | None = None,
        extract: "TrueFalseT" = True,
        force_refresh: bool = False,
    ) -> str | bytes:
        """Get Activation bytes from Audible.

        Args:
            filename: [Optional] filename to save the activation blob
            extract: [Optional] if True, returns the extracted activation
                bytes otherwise the whole activation blob
            force_refresh: [Optional] if True, existing activation bytes in
                auth file will be ignored and new activation bytes
                will be requested from server.

        Returns:
            The activation bytes

        .. versionadded:: 0.5.1
           The ``force_refresh`` argument. Fetched activation bytes are now
           stored to `Authententicator.activation_bytes`.
        """
        if not force_refresh and extract and self.activation_bytes is not None:
            logger.debug("Activation bytes already fetched. Returned saved one.")
            return self.activation_bytes

        logger.debug("Fetch activation blob from server now.")
        ab = get_ab(self, filename, extract)  # type: ignore[arg-type]

        if extract:
            logger.debug(
                "Extract activation bytes from blob and store value"
                "activation_bytes attribute."
            )
            logger.debug("Found activation bytes: %s", ab)
            self.activation_bytes = ab

        return ab

    def user_profile(self) -> dict[str, Any]:
        if self.access_token is None:
            raise Exception("No access token found.")
        if self.locale is None:
            raise Exception("No locale found.")

        return user_profile(access_token=self.access_token, domain=self.locale.domain)

    @property
    def access_token_expires(self) -> timedelta:
        if self.expires is None:
            raise Exception("No expires timestamp found.")
        return datetime.fromtimestamp(self.expires, timezone.utc) - datetime.now(
            timezone.utc
        )

    @property
    def access_token_expired(self) -> bool:
        if self.expires is None:
            raise Exception("No expires timestamp found.")
        return datetime.fromtimestamp(self.expires, timezone.utc) <= datetime.now(
            timezone.utc
        )
